/******************************************************************************
MODULE: huffman.cpp
PROEJCT: huffman
PROGRAMMER: Jeff Nielsen
DATE: February 1996

DESCRIPTION: This module implements a Huffman tree demo, annotated with SAIL
	     function calls.

USAGE: huffman <frequency_filename>

******************************************************************************/
#include <iostream.h>
#include <iomanip.h>
#include <stdio.h>
#include <fstream.h>
#include <string.h>
#include <strstrea.h>

#define TRUE 1
#define FALSE 0
#define UNDEF -1
#define SWAN

#include "heap.ht"	// Heap template
#include "sail.h"


// Huffman class ------------------------------------------------------------

struct HuffNode {	// Base class for Internal & External nodes
   int freq;		// Frequency value
   virtual bool is_leaf() const = 0;
};

class HuffTree {	// Basically a pointer to a Huffman node, with methods
private:
   HuffNode* root;	// Pointer to root node
   bool newline;        // For use with printing routine
public:
   HuffTree() {};
   HuffTree(HuffNode* rootnode) { root = rootnode; }
   HuffNode* getroot() { return root; }
   int val() { return root->freq; }
   bool operator < (HuffTree& other) { return val() < other.val(); }
   bool operator > (HuffTree& other) { return val() > other.val(); }
   bool operator == (HuffTree& other) { return val() == other.val();}
   void printhelp(const HuffNode*, const int);
   void print() { printhelp(root, 0); cout << endl; }
#ifdef SWAN
   void swan_drawhelp(HuffNode*, GRAPHID);
   void swan_draw(GRAPHID g) { swan_drawhelp(root, g); }
#endif
};

struct HuffIntNode : public HuffNode {	  // Internal Huffman node
   HuffNode *left, *right;
   HuffIntNode(HuffNode* l, HuffNode* r)
      { left = l; right = r; freq = l->freq + r->freq; }
   HuffIntNode(HuffTree l, HuffTree r) {
      left = l.getroot(); right = r.getroot();
      freq = left->freq + right->freq;
   }
   virtual bool is_leaf() const { return FALSE; }
};

struct HuffLeafNode : public HuffNode {	  // Leaf Huffman node
   char c;
   HuffLeafNode(char ch, int frequency) { c = ch; freq = frequency; }
   virtual bool is_leaf() const { return TRUE; }
};

//////////////////////////////////////////////////
void HuffTree::printhelp(const HuffNode* rt, const int level)
//
// PURPOSE: Recursive function for printing out the tree.  Prints a pre-order
//          traversal, with left children printed on same line as parent &
//          right children printed on following line.  Values are printed at
//          column 5 * level.
// PARAMETERS:
//    rt: Pointer to current node
//    level: Tree level of the current node */
{
   if (newline) for (int i = 0; i < level; i++) cout << "     ";
   cout << setiosflags(ios::left) << setw(3) << rt->freq;
   newline = FALSE;

   if (rt->is_leaf()) cout << ((HuffLeafNode*)rt)->c << ' ';
   else {
      cout << "  ";
      printhelp(((HuffIntNode*)rt)->right, level + 1);   // Print left subtree
      if (!newline) { cout << '\n'; newline = TRUE; }	 // Move down a line
      printhelp(((HuffIntNode*)rt)->left, level + 1);    // Print right subtree
   }
}

#ifdef SWAN

//////////////////////////////////////////////////
void HuffTree::swan_drawhelp(HuffNode* node_ptr, GRAPHID g)
//
// PURPOSE: Recursive function for drawing the representation of the Huffman
//	    tree in SWAN.
// PARAMETERS:
//    node_ptr: Pointer to the "root" node of the tree being drawn
//    g: Graph in which to insert node
{
   char memstr[10];
   ostrstream memstm(memstr, 10);
   if (node_ptr->is_leaf()) {
      memstm.seekp(0);
      memstm << " " << ((HuffLeafNode*)node_ptr)->c << ": " << node_ptr->freq
	     << ends;
      sw_insertnode(g, NODEID(node_ptr), memstr);
      sw_setnodeattr(g, NODEID(node_ptr), GNODELINETH, THICKLINE);
      sw_setnodeattr(g, NODEID(node_ptr), GNODEWIDTH, 46);
   }
   else {         // It's an internal node
      sw_insertnode(g, NODEID(node_ptr), inttostr(node_ptr->freq));
      sw_setnodeattr(g, NODEID(node_ptr), GNODETYPE, CIRCLE);
      sw_setnodeattr(g, NODEID(node_ptr), GNODEWIDTH, 35);
      HuffNode* leftchild = ((HuffIntNode*)node_ptr)->left;
      HuffNode* rightchild = ((HuffIntNode*)node_ptr)->right;

      // Now insert children--don't need to check if they exist (as in other
      // binary trees) because all Huffman trees are full
      swan_drawhelp(leftchild, g);       // call recursively
      sw_insertbinedge(g, NODEID(node_ptr), NODEID(leftchild), "0",
                        LEFTCHILD);
      swan_drawhelp(rightchild, g);      // call recursively
      sw_insertbinedge(g, NODEID(node_ptr), NODEID(rightchild), "1",
                        RIGHTCHILD);
   }
}
#endif

// CodeTable class --------------------------------------------------------
class CodeTable {
private:
   struct {
      char c;
      char code[20];
   } table[50];		// Holds the codes
   int num_codes;	// How many entries in the table
   char prefix[20];     // Stores codes being built

   void readcodes(HuffNode*, int);
   int findcode(char);
public:
   CodeTable(HuffTree& ht)
      { num_codes = 0; readcodes(ht.getroot(), 0); }
   bool encode(char*, char*);
   void print();
};

//////////////////////////////////////////////////
void CodeTable::readcodes(HuffNode* node, int level)
//
// PURPOSE: Recursive function to build the code table from a Huffman tree
// PARAMETERS:
//    node: Current node being examined
//    level: Current depth in the tree
{
   if (node->is_leaf()) {
      table[num_codes].c = ((HuffLeafNode*)node)->c;
      strcpy(table[num_codes++].code, prefix);
   }
   else {
      prefix[level] = '0';
      prefix[level+1] = '\0';
      readcodes(((HuffIntNode*)node)->left, level+1);
      prefix[level] = '1';
      prefix[level+1] = '\0';
      readcodes(((HuffIntNode*)node)->right, level+1);
      prefix[level] = '\0';
   }
}

//////////////////////////////////////////////////
void CodeTable::print()
// PURPOSE: Print out the code table
{
   int total_lengths = 0;
   cout << "CODE TABLE\n----------\n";
   for (int i = 0; i < num_codes; i++) {
      cout << table[i].c << ": " << table[i].code << endl;
      total_lengths += strlen(table[i].code);
   }
}

//////////////////////////////////////////////////
int CodeTable::findcode(char c)
//
// PURPOSE: Find the location of a letter/code pair in the table
// PARAMETERS:
//    c: The character to find
// RETURNS: The index of the pair in the table, or UNDEF if not found
{
   for (int i = 0; i < num_codes; i++) if (table[i].c == c) return i;
   return UNDEF;
}

//////////////////////////////////////////////////
bool CodeTable::encode(char *dest, char *src)
//
// PURPOSE: Encode a word using the Huffman codes stored in the table
// PARAMETERS:
//    dest: String in which to write the code
//    src: String holding the word to decode
// RETURNS: TRUE if the word was successfully encoded, FALSE if not
{
   for (int i = 0; i < strlen(src); i++) {
      int pos = findcode(src[i]);
      if (pos == UNDEF) {
#ifndef SWAN
         cout << "ERROR: No code for character " << src[i] << endl;
#endif
         return FALSE;
      }
      else {
         strcpy(dest, table[pos].code);
         dest += strlen(table[pos].code);
      }
   }
   return TRUE;
}

//////////////////////////////////////////////////
bool decode(char *dest, char *src, HuffTree& ht)
//
// PURPOSE: Decode a bit-string using a Huffman tree--not a member of class
//	    CodeTable, but included here because it's a similar function.
// PARAMETERS:
//    dest: String in which to write the decoded word
//    src: String holding the bit-string to be decoded
//    ht: Huffman coding tree to use for the decoding
// RETURNS: TRUE if a valid bit-string was entered, FALSE if not
{
   int dest_count = 0;
   HuffNode* node = ht.getroot();
   while (*src != '\0') {
      if (*src == '0') node = ((HuffIntNode*)node)->left;
      else if (*src == '1') node = ((HuffIntNode*)node)->right;
      else {
#ifndef SWAN
	 cout << "ERROR: Non-binary digit " << *src << " encountered.\n";
#endif
	 return FALSE;
      }
      if (node->is_leaf()) {
         dest[dest_count++] = ((HuffLeafNode*)node)->c;
         node = ht.getroot();
      }
      src++;
   }
   dest[dest_count] = '\0';      // Null-terminate
   return TRUE;
}

// Tree building/displaying functions -----------------------------------------

typedef MinHeap<HuffTree> HuffHeap;

#ifdef SWAN
int num_graphs = 0;

//////////////////////////////////////////////////
void erase_trees() {
   for (int i = 1; i <= num_graphs; i++) sw_deletegraph(GRAPHID(i));
   num_graphs = 0;   // Remove any existing graphs
}

//////////////////////////////////////////////////
void draw_onetree(HuffTree tree) {
   GRAPHID gid = sw_newgraph(GRAPHID(++num_graphs), BOX, UNDIRECTED);
   LC_ID lc = sw_newlc(gid, NULLLCID, TREE);
   sw_setcurlc(gid, lc);
   tree.swan_draw(gid);
}

//////////////////////////////////////////////////
void draw_trees(HuffHeap &h)
//
// PURPOSE: Draw all of the Huffman trees currently in the heap
// PARAMETERS:
//    h: The heap--reference variable so you don't pass the whole thing
{
   for (int i = 0; i < h.size(); i++) draw_onetree(h.get(i));
   sw_displayallgraphs();
}

#endif

//////////////////////////////////////////////////
HuffTree huffman(HuffTree nodearray[], int n)
//
// PURPOSE: Builds a Huffman Coding Tree.  This code is based on p.340 of
//     Cormen, et al. Introduction to Algorithms
// PARAMETERS:
//    nodearray: Array of character/frequency pairs each stored in a
//               HuffTree consisting of a single HuffLeafNode
//    n: Number of character/frequency pairs
{
   HuffHeap q(nodearray, n);      // Make nodearray into a heap
   q.buildheap();                 // Build the heap
#ifdef SWAN
   draw_trees(q);
   sw_print("Here are the leaf nodes.  Press STEP to continue.");
   sw_break();
#endif

   HuffIntNode *z;
   for (int i = 0; i < n-1; i++) {
      z = new HuffIntNode(q.removemin(), q.removemin());
#ifdef SWAN
      erase_trees();
      draw_onetree(HuffTree(z));
      draw_trees(q);
      sw_print("The two trees with lowest frequency have been combined.");
      sw_break();
#endif
      q.insert(HuffTree(z));
#ifdef SWAN
      if (q.size() > 1) {
         erase_trees();
         draw_trees(q);
         sw_print("The new tree has been re-inserted into the list.");
         sw_break();
      }
#endif
   }      // End for
   return q.removemin();
}


// Main functions -----------------------------------------------------------

//////////////////////////////////////////////////
void do_commands(HuffTree& ht, CodeTable& ct)
//
// PURPOSE: Wait for encode/decode commands from the user and execute them.
// PARAMETERS:
//    ht: The Huffman tree with which to read the codes
//    ct: The code table from which to get the letter
{
   char output[500], result[500];
#ifdef SWAN
   char memstr[500];	// In-memory string space
   char* ptr;   
   while (TRUE) {
      sw_print("Enter \"e <word>\" to encode or \"d <bit-string>\" to decode.");
      sw_getstr(memstr);
      if (memstr[0] == 'e') {
	    ptr = &(memstr[1]);
		while(*ptr == ' ') ptr++;
	 if (ct.encode(result, ptr))
		sprintf(output, "The code for |%s| is %s.       Press STEP or QUIT.",
		   ptr, result);
	 else sprintf(output,"ERROR: Character entered for which no code exists.       Press STEP or QUIT.");
      }
      else if (memstr[0] == 'd') {
	    ptr = &(memstr[1]);
		while(*ptr == ' ') ptr++;

	 if (decode(result, ptr, ht))
	    sprintf(output,"The word |%s| is: %s.       Press STEP or QUIT.", ptr, result);
	 else sprintf(output,"ERROR: Non-binary digit encountered.       Press STEP or QUIT.");
      }
      else  sprintf(output,"Syntax error on entry--begin with 'e' or 'd'.       Press STEP or QUIT.");

      sw_print(output);
      sw_break();
   }
#else
   char cmd;
   cin.clear();
   cout << "Now enter words to encode (e) or bit strings to decode (d):"
        << endl;
   cin >> cmd >> instr;
   while (cin) {
      if (cmd == 'e' && ct.encode(result, instr)) {
         int length = strlen(result);
         cout << "The code is " << result << "\n  (" << length
              << " bits or " << float(length) / (8.0 * strlen(instr))
              << " the size of 8-bit encoding).\n";
      }
      else if (cmd == 'd' && decode(result, instr, ht))
            cout << "The word is: " << result << endl;
      else if (cmd != 'e' && cmd != 'd')
         cout << "Syntax error--must begin with either 'e' or 'd'" << endl;
      cin >> cmd >> instr;
   }
#endif
}

//////////////////////////////////////////////////
int GS_MAIN(GS_MAIN_ARGS_F)
{
   char ch;
   int freq;
   HuffTree treearray[50], final_tree;
   int arrsize = 0;
   char    temp_str[80] ;

   sw_init(GS_MAIN_ARGS_A) ;
   sw_step();

   sw_print("What file would you like to use?") ;
   sw_getstr(temp_str) ;
   ifstream fs(temp_str);
   if (!fs) { sw_print("Couldn't open file "); exit(-1); }

   fs >> ch >> freq;      // Read in character/frequency pairs
   while (fs) {
      treearray[arrsize++] = HuffTree(new HuffLeafNode(ch, freq));
      fs >> ch >> freq;
   }
   if (arrsize == 0)
      { cout << "ERROR: No character-frequency pairs read.\n"; exit(-1); }


   final_tree = huffman(treearray, arrsize);

#ifdef SWAN
   sw_disablebuttons(RUN);    // Make them step through the last part
   sw_step();		      // In case they have set it to RUN mode
   sw_print("This is the final Huffman tree.  Press STEP to continue.");
   sw_break();
#else
   final_tree.print();
#endif
   CodeTable codes(final_tree);        // Construct a code table from the tree
   do_commands(final_tree, codes);     // Allow user to decode & encode
   return 0;
}













