/*****************************************************************************
MODULE: redbdriv.cpp
PROJECT: redblack
PROGRAMMER: Jeff Nielsen
DATE: February 1996

DESCRIPTION: Driver module for the programs.  This program constructs and
             maintains a Red-Black tree, annotated with SAIL function calls to
	     be able to visualize the tree.  Compile with -DSWAN for the
	     annotated version.

******************************************************************************/

// Header Files ---------------------------------------------------------------
#define SWAN
#include <iostream.h>
#include <strstrea.h>
#include <iomanip.h>
#ifndef SWAN
#include <ctype.h>
#endif
#include "redblack.h"
#include "stdjeff.h"

// Symbolic Constants & Typedefs ---------------------------------------------
#define INSERT 105	      // Constants for the non-annotated version
#define REMOVE 114
//#define FIND 102	      // Find command disabled in this version
#define LIST 108
#define MAKENULL 109
#define QUIT 113

#define BUFLENGTH 162	      // Length of I/O buffers

// Swan functions ----------------------------------------------------------
#ifdef SWAN
#include "sail.h"
GRAPHID swangraph = GRAPHID(100);   // Name for graph in Swan window
extern rb_node *NULLNODE;	    // Represents NULL node in tree

//////////////////////////////////////////////////
inline void make_swangraph(GRAPHID gid, NODETYPE nt, bool dir_undir,
            LAYOUTCOMPTYPE lctype)
//
// PURPOSE: Macro for making SWAN graphs
// PARAMETERS:
//    gid: The graphid you want for this graph
//    nt: The type of node you want the graph to have
//    dir_undir: DIRECTED or UNDIRECTED graph
//    lctype: The default layout component you want this graph to have.
{
   sw_newgraph(gid, nt, dir_undir);
   sw_newlc(gid, LC_ID(1), lctype);
   sw_setcurlc(gid, LC_ID(1));
}

//////////////////////////////////////////////////
void swtree_insert(rb_node* node_ptr)
//
// PURPOSE: Recursive function for inserting nodes & edges in the Swan tree
//	    representation
// PARAMETERS:
//    node_ptr: Pointer to the current node
{
    rb_node *left_child ;
    rb_node *right_child ;

    if (node_ptr == NULLNODE) return;
    sw_insertnode(swangraph, NODEID(node_ptr), inttostr(node_ptr->key));
    sw_setnodeattr(swangraph, NODEID(node_ptr), GNODECOLOR,
		   (node_ptr->color == RED ? MAGENTA : MGRAY));
    if ((left_child = node_ptr->left) != NULLNODE) {
        swtree_insert(left_child) ;
        sw_insertbinedge(swangraph, NODEID(node_ptr), NODEID(left_child), 
		         NULL, LEFTCHILD) ;
    }
    if ((right_child = node_ptr->right) != NULLNODE) {
        swtree_insert(right_child) ;
        sw_insertbinedge(swangraph, NODEID(node_ptr), NODEID(right_child), 
		         NULL, RIGHTCHILD) ;
    }
}

//////////////////////////////////////////////////
void rebuild_tree(redblack_tree* tree_ptr)
//
// PURPOSE: Function to redraw the red-black tree in the Swan window.  Deletes
//	    the existing tree drawing and calls the recursive swtree_insert()
//	    function.
// PARAMTERS:
//    tree_ptr: Pointer to the tree object you wish to draw
{
    static int count;     // 0 if this function hasn't been called yet
    if (count++ > 0) sw_deletegraph(swangraph);	   // Delete existing graph

    //    make_swangraph(swangraph, CIRCLE, UNDIRECTED, BINTREE);
    // NOTE: Using my make_swangraph macro causes a bug in this case--
    // for some reason if you try to use sw_setgraphattr after using
    // sw_newlc, it doesn't work.  Weird!!
    
    sw_newgraph(swangraph, CIRCLE, UNDIRECTED);
    sw_setgraphattr(swangraph, GNODEFILLED, TRUE);
    sw_setgraphattr(swangraph, GEDGELABEL, OFF);
    sw_newlc(swangraph, LC_ID(1), BINTREE);
    sw_setcurlc(swangraph, LC_ID(1));
    swtree_insert(tree_ptr->get_root());
    sw_displayallgraphs();
}

#endif

//////////////////////////////////////////////////
int GS_MAIN(GS_MAIN_ARGS_F)
//
// PURPOSE: Driver for program
{
   redblack_tree tree;
   bool done = FALSE;

#ifdef SWAN
   sw_init(GS_MAIN_ARGS_A);
   //   sw_disablebuttons(RUN);
   sw_enablemenuitem(EDITMENU, ITEMINSNODE);
   sw_enablemenuitem(EDITMENU, ITEMDELNODE);
   sw_step();
   sw_print("Use the Edit menu to work with the red-black tree.");
#endif
   
   do {
      ELEM key;

#ifdef SWAN
      char instr[300];
      istrstream instm(instr, 300);		     // Attach stream

      switch(sw_wait(STEP | RUN | INSNODE | DELNODE)) {
	 case STEP:
	    sw_step();
	    sw_print("The program will run in STEP mode.");
	    break;

	 case RUN:
	    sw_run();
	    sw_print("The program will run without breakpoints.");
	    break;

	 case INSNODE:
	    sw_print("Enter a value to be inserted:");
	    sw_getstr(instr);
	    instm.seekg(0);
	    instm >> key;
	    tree.insert(key);
	    rebuild_tree(&tree);
	    break;

	 case DELNODE:
	    NODEID n;
	    if (!tree.is_empty()) {
	       sw_print("Click on a node to remove.");
	       sw_picknode(&n);
	       sw_getnodeattr(swangraph, n, GNODELABEL, LABEL(instr));
	       instm.seekg(0);
	       instm >> key;
	       tree.remove(key);
	       rebuild_tree(&tree);
	    }
	    else sw_print("Can't remove any nodes--the tree is empty.");
	    break;
      }
#else      
      char command_line[BUFLENGTH], cmd[21];

      cout << "\n==>";
      if (cin.getline(command_line, BUFLENGTH)) {
         cout << command_line << endl;
	 istrstream cmdstream(command_line); // Attach command_line to stream
         cmdstream >> cmd;                   // Get the first word

         switch(tolower(cmd[0])) {
            case INSERT:
               cmdstream >> key;
               tree.insert(key);
               break;

            case REMOVE:
               cmdstream >> key;
               tree.remove(key);
               break;

//          case FIND:
//             cmdstream >> key;
//             tree.find(key);
//             break;

            case LIST:
               tree.print();
               break;

            case MAKENULL:
               tree.clear();
               break;

            case QUIT:
               done = TRUE;
               break;

            default:
               cout << "Invalid input.";
         }                                   // end switch
      }                                      // end if (cin.getline())
      else { done = TRUE; cout << endl; }    // Reached EOF
#endif

   } while (!done);                          // end do
   return 0;
}













